/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* namecapture.c *** X Program to Capture Process Variable Names */

#include "namecaptureVersion.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <X11/cursorfont.h>
#include <X11/Intrinsic.h>
#include <X11/Xatom.h>
#include <Xm/Xm.h>
#include <Xm/CascadeB.h>
#include <Xm/CascadeBG.h>
#include <Xm/DragDrop.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/MainW.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrolledW.h>
#include <Xm/SelectioB.h>
#include <Xm/TextF.h>
#include <Xm/Text.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <Xm/Separator.h>

/* WIN32 differences */
#ifdef WIN32
#include <windows.h>
/* Hummingbird extra functions including lprintf
 *   Needs to be included after Intrinsic.h for Exceed 5
 *   (Intrinsic.h is included in xtParams.h) */
# include <X11/XlibXtra.h>
/* Needed for access */
# include <io.h>
/* This is done in Exceed 6 but not in Exceed 5
 *   Need it to define printf as lprintf for Windows
 *   (as opposed to Console) apps */
# ifdef _WINDOWS
#  ifndef printf
#   define printf lprintf
#  endif    
# endif    
#else /* #ifdef WIN32 */
/* WIN32 does not have unistd.h */
# include <unistd.h>
/* WIN32 does not have utsname.h */
# include <sys/utsname.h>
#endif /* #ifdef WIN32 */

#ifdef EDITRES
#include <X11/Xmu/Editres.h>
#endif

#define SDDSID "SDDS1"
 /* Expect up to 18 PV names from MEDM */
#define TEXTMAX 1024
#define CANAMESIZE 40
#define MAXARGS 10
#define PATTERN "*"
#define DIRECTORY ""
#define TITLE "Name Capture"

/* Function prototypes */

int main(int argc, char **argv);
void dragdropfinishcb(Widget w, XtPointer clientdata, XtPointer calldata);
void dropproc(Widget w, XtPointer clientdata, XtPointer calldata);
void eventtext(Widget w, XtPointer clientdata, XEvent *event,
  Boolean *ctd);
void filecb(Widget w, XtPointer clientdata,  XtPointer calldata);
int getuname(void);
void helpcb(Widget w, XtPointer clientdata, XtPointer calldata);
void killwidget(Widget w, XtPointer clientdata, XtPointer calldata);
void quit(Widget w, XtPointer clientdata, XtPointer calldata);
void textcb(Widget w, XtPointer clientdata, XtPointer calldata);
void timerproc(XtPointer clientdata, XtIntervalId *id);
Boolean workproc(XtPointer clientdata);
void writeburtheader(FILE *file);
void xerrmsg(char *fmt, ...);
void xinfomsg(char *fmt, ...);
int yesno(Widget wparent, char * question);
void yesnoresponse(Widget w, XtPointer clientdata, XtPointer calldata);

/* Structures */

/* Global variables */

XtAppContext appcontext;
Display *display;
int screen;
Window rootwindow;
Widget appshell,mainwin,mainmenu,mainform,warningbox,infobox;
Widget helppopup,text;
Pixmap apspix;
Arg args[MAXARGS];
Cardinal nargs;

char string[BUFSIZ];
XmString cstring;
int windowmessage=0;
int burtmode=1,autofile=1;

char filename[BUFSIZ]="";
char savepattern[BUFSIZ]=PATTERN;
char savedirectory[BUFSIZ]=DIRECTORY;
FILE *file=(FILE *)0;

char *unamestring;

void (*olddropproc)();

/* Define icons (Refer to as name_bits[]) */

#include "aps.icon"

/* Define help text */

#include "namecapture.hlp"

/**************************** main ****************************************/
int main(int argc, char **argv)
{
    Widget w,w1;

  /* WIN32 initialization */
#ifdef WIN32	
    HCLXmInit();
#endif

  /* Constants and defaults */
    getuname();

  /* Set up top level */
    XtToolkitInitialize();
    appcontext=XtCreateApplicationContext();

  /* Get display parameters */
    display=XtOpenDisplay(appcontext,NULL,argv[0],"NameCapture",NULL,0,&argc,argv);
    if(display == NULL) {
	xerrmsg("Namecapture initialization: Cannot open display");
	exit(1);
    }
    screen=DefaultScreen(display);
    rootwindow=RootWindow(display,screen);

  /* Set icon */
    apspix=XCreatePixmapFromBitmapData(display,rootwindow,
      (char *)aps_bits,aps_width,aps_height,
      1,0,1);
    nargs=0;
#ifndef WIN32    
    XtSetArg(args[nargs],XtNiconPixmap,apspix); nargs++;
#endif
    XtSetArg(args[nargs],XmNtitle,TITLE); nargs++; 
    appshell=XtAppCreateShell("namecapture","ControlApp",
      applicationShellWidgetClass,display,args,nargs);

  /* Enable Editres */
#ifdef EDITRES
    XtAddEventHandler(appshell,(EventMask)NULL,True,
      (XtEventHandler)_XEditResCheckMessages,NULL);
#endif    

  /* Define main window */
    nargs=0;
    mainwin=XmCreateMainWindow(appshell,"mainWindow",args,nargs);
    XtManageChild(mainwin);

  /* Define menu bar */
    nargs=0;
    mainmenu=XmCreateMenuBar(mainwin,"mainMenu",args,nargs);
    XtManageChild(mainmenu);

  /* Define file button */
    nargs=0;
    XtSetArg(args[nargs],XmNtearOffModel,XmTEAR_OFF_ENABLED); nargs++;
    w=XmCreatePulldownMenu(mainmenu,"fileMenu",args,nargs);

    nargs=0;
    XtSetArg(args[nargs],XmNmnemonic,'N'); nargs++;
    w1=XmCreatePushButton(w,"Open New/Reuse",args,nargs);
    XtManageChild(w1);
    XtAddCallback(w1,XmNactivateCallback,filecb,(XtPointer)1);

    nargs=0;
    XtSetArg(args[nargs],XmNmnemonic,'O'); nargs++;
    w1=XmCreatePushButton(w,"Open Append",args,nargs);
    XtManageChild(w1);
    XtAddCallback(w1,XmNactivateCallback,filecb,(XtPointer)2);

    nargs=0;
    XtSetArg(args[nargs],XmNmnemonic,'W'); nargs++;
    w1=XmCreatePushButton(w,"Write            (Btn3)",args,nargs);
    XtManageChild(w1);
    XtAddCallback(w1,XmNactivateCallback,filecb,(XtPointer)3);

    nargs=0;
    XtSetArg(args[nargs],XmNmnemonic,'C'); nargs++;
    w1=XmCreatePushButton(w,"Close",args,nargs);
    XtManageChild(w1);
    XtAddCallback(w1,XmNactivateCallback,filecb,(XtPointer)4);

    nargs=0;
    XtSetArg(args[nargs],XmNmnemonic,'B'); nargs++;
    w1=XmCreateToggleButton(w,"BURT Mode",args,nargs);
    XtManageChild(w1);
    XtAddCallback(w1,XmNvalueChangedCallback,filecb,(XtPointer)5);
    XmToggleButtonSetState(w1,(Boolean)burtmode,False);
    
    nargs=0;
    XtSetArg(args[nargs],XmNmnemonic,'A'); nargs++;
    w1=XmCreateToggleButton(w,"Automatic File Insert",args,nargs);
    XtManageChild(w1);
    XtAddCallback(w1,XmNvalueChangedCallback,filecb,(XtPointer)6);
    XmToggleButtonSetState(w1,(Boolean)autofile,False);
    
    nargs=0;
    w1=XmCreateSeparator(w,"separator",args,nargs);
    XtManageChild(w1);
    
    nargs=0;
    XtSetArg(args[nargs],XmNmnemonic,'Q'); nargs++;
    w1=XmCreatePushButton(w,"Quit",args,nargs);
    XtManageChild(w1);
    XtAddCallback(w1,XmNactivateCallback,quit,(XtPointer)NULL);

    nargs=0;
    XtSetArg(args[nargs],XmNsubMenuId,w); nargs++;
    XtSetArg(args[nargs],XmNmnemonic,'F'); nargs++;
    w1=XmCreateCascadeButton(mainmenu,"File",args,nargs);
    XtManageChild(w1);

  /* Define help button */
    nargs=0;
    XtSetArg(args[nargs],XmNtearOffModel,XmTEAR_OFF_ENABLED); nargs++;
    w=XmCreatePulldownMenu(mainmenu,"helpMenu",args,nargs);

    nargs=0;
    XtSetArg(args[nargs],XmNmnemonic,'O'); nargs++;
    w1=XmCreatePushButton(w,"Overview",args,nargs);
    XtManageChild(w1);
    XtAddCallback(w1,XmNactivateCallback,helpcb,(XtPointer)0);

    nargs=0;
    XtSetArg(args[nargs],XmNmnemonic,'V'); nargs++;
    w1=XmCreatePushButton(w,"Version",args,nargs);
    XtManageChild(w1);
    XtAddCallback(w1,XmNactivateCallback,helpcb,(XtPointer)1);

    nargs=0;
    XtSetArg(args[nargs],XmNsubMenuId,w); nargs++;
    XtSetArg(args[nargs],XmNmnemonic,'H'); nargs++;
    w1=XmCreateCascadeButton(mainmenu,"Help",args,nargs);
    XtManageChild(w1);

    nargs=0;
    XtSetArg(args[nargs],XmNmenuHelpWidget,w1); nargs++;
    XtSetValues(mainmenu,args,nargs);

  /* Define form for rest of main window */
    nargs=0;
    mainform=XmCreateForm(mainwin,"form",args,nargs);
    XtManageChild(mainform);

  /* Define text box */    
    nargs=0;
    XtSetArg(args[nargs],XmNmaxLength,TEXTMAX); nargs++;
    XtSetArg(args[nargs],XmNcolumns,CANAMESIZE); nargs++;
    text=w=XmCreateTextField(mainform,"mainText",args,nargs);
    XtManageChild(w);
    XtAddEventHandler(text,ButtonPressMask,False,eventtext,NULL);

  /* Retrieve current drop site resources */
    nargs=0;
    XtSetArg(args[nargs],XmNdropProc,&olddropproc); nargs++;
    XmDropSiteRetrieve(w,args,nargs);

  /* Update drop site resources */
    nargs=0;
    XtSetArg(args[nargs],XmNdropProc,dropproc); nargs++;
    XmDropSiteUpdate(w,args,nargs);

  /* Set main window areas */
    XmMainWindowSetAreas(mainwin,mainmenu,NULL,NULL,NULL,mainform);

  /* Realize */
    XtRealizeWidget(appshell);

  /* Start */
    windowmessage=1;
    XtAppMainLoop(appcontext);
    return(0);
}
/**************************** dragdropfinishcb  ***************************/
#ifdef TEST
void dragdropfinishcb(Widget w, XtPointer clientdata, XtPointer calldata)
{
    XmTextPosition len;
    XmDragDropFinishCallback dropdata=(XmDragDropFinishCallback)calldata;

    printf("text=%d w=%d \n",text,w);

  /* Select the new text */
    len=XmTextGetLastPosition(text);
  /*    printf("DragDropFinishCallback Last Position: |%d|\n",len); */
    XmTextSetSelection(text,0,len,dropdata->timeStamp);
}
#endif
/**************************** dropproc ************************************/
void dropproc(Widget w, XtPointer clientdata, XtPointer calldata)
{
    XmTextPosition len;
    XmDropProcCallback dropdata=(XmDropProcCallback)calldata;

  /* Delete the text that is there */    
    if(dropdata->dropAction == XmDROP || dropdata->dropAction == XmDROP_COPY) {
	len=XmTextGetLastPosition(w);
	XmTextSetSelection(w,0,len,dropdata->timeStamp);
	XmTextRemove(w);
    }

  /* Add a dragDropFinishCallback */
  /*     XtAddCallback(dropdata->dragContext,XmNdragDropFinishCallback, */
  /*       dragdropfinishcb,(XtPointer)NULL); */
    
  /* Add a workproc */
  /*     XtAppAddWorkProc(appcontext,workproc,NULL); */
    
  /* Add a timer proc (time in msec) */
    XtAppAddTimeOut(appcontext,200,timerproc,NULL);

  /* Call the old procedure */    
    (*olddropproc)(w,clientdata,calldata);
}
/**************************** eventtext ***********************************/
void eventtext(Widget w, XtPointer clientdata, XEvent *event,
  Boolean *ctd)
{
    switch (event->type) {
    case ButtonPress:
	switch(event->xbutton.button) {
	case Button3:
	    filecb(w,(XtPointer)3,(XtPointer)0);
	    break;
	}
    default:
	break;
    }
}
/**************************** filecb **************************************/
void filecb(Widget w, XtPointer clientdata,  XtPointer calldata)
{
    static Widget wdialog;
    int n;
    static int nsave=-1;
    XmString directory,pattern;
    char *value;
    XmFileSelectionBoxCallbackStruct *cbs;

    n=(int)clientdata;
    cbs=(XmFileSelectionBoxCallbackStruct *)calldata;
    switch(n) {
    case 1:     /* Open New/Reuse*/
    case 2:     /* Open Append */
	pattern=XmStringCreateLocalized(savepattern);
	directory=XmStringCreateLocalized(savedirectory);
	nargs=0;
	XtSetArg(args[nargs],XmNtitle,"Namecapture Filename"); nargs++;
	XtSetArg(args[nargs],XmNpattern,pattern); nargs++;
	XtSetArg(args[nargs],XmNdirectory,directory); nargs++;
	wdialog=XmCreateFileSelectionDialog(mainwin,"fileSelection",
	  args,nargs);
	XmStringFree(pattern);
	XmStringFree(directory);
	XtAddCallback(wdialog,XmNcancelCallback,killwidget,(XtPointer)0);
	XtAddCallback(wdialog,XmNokCallback,filecb,(XtPointer)100);
	XtSetSensitive(XmSelectionBoxGetChild(wdialog,
	  XmDIALOG_HELP_BUTTON),False);
	XtManageChild(wdialog);
	nsave=n;
	break;
    case 3:     /* Write */
	if(file) {
	    value=XmTextFieldGetString(text);
	    if(burtmode) fprintf(file,"%s pv RO\n",value);
	    else fprintf(file,"%s\n",value);
	    fflush(file);
	    XFree(value);
	    XBell(display,50); 	    
	}
	else xerrmsg("No file is open");
	break;
    case 4:     /* Close */
	if(file) {
	    fclose(file);
	    *filename='\0';
	    file=(FILE *)0;
	    nargs=0;
	    XtSetArg(args[nargs],XmNtitle,TITLE); nargs++;
	    XtSetValues(appshell,args,nargs);
	}
	else xerrmsg("No file is open");
	break;
    case 5:
	burtmode=XmToggleButtonGetState(w)?1:0;
	break;
    case 6:
	autofile=XmToggleButtonGetState(w)?1:0;
	break;
    case 100:
	if(!XmStringGetLtoR(cbs->value,XmSTRING_DEFAULT_CHARSET,&value)) {
	    xerrmsg("Error getting filename");
	    return;
	}
	if(!*value) {
	    xerrmsg("File name is blank");
	    XtFree(value);
	    return;
	}
	strcpy(filename,value);
	XtFree(value);
      /* Get the patterns the user left */
	if(XmStringGetLtoR(cbs->dir,XmSTRING_DEFAULT_CHARSET,&value)) {
	    strcpy(savedirectory,value);
	    XtFree(value);
	}
	if(XmStringGetLtoR(cbs->pattern,XmSTRING_DEFAULT_CHARSET,&value)) {
	    strcpy(savepattern,value);
	    XtFree(value);
	}
	XtDestroyWidget(wdialog);
      /* Do what was requested */
	if(file) {
	    fclose(file);
	    *filename='\0';
	    file=(FILE *)0;
	    nargs=0;
	    XtSetArg(args[nargs],XmNtitle,TITLE); nargs++;
	    XtSetValues(appshell,args,nargs);
	}
	switch(nsave) {
	case 1:     /* Open New/Reuse */
	    if(access(filename,0) == 0) {
		if(yesno(appshell,"File already exists, Reuse?") != 1)
		  return;
	    }
	    file=fopen(filename,"w");
	    if(!file) {
		xerrmsg("Could not open % for writing",filename);
		break;
	    }
	    if(burtmode) writeburtheader(file);
	    break;
	case 2:     /* Open Append */
	    file=fopen(filename,"a");
	    if(!file) xerrmsg("Could not open % for appending",filename);
	    else {
		value=XmTextFieldGetString(text);
		if(burtmode) fprintf(file,"%s pv RO\n",value);
		else fprintf(file,"%s\n",value);
		fflush(file);
		XFree(value);
	    }
	    break;
	}
	if(file) {
	    char *ptr;
	    
	    ptr=strrchr(filename,'/');
	    if(!ptr) ptr=filename;
	    else ptr++;
	    nargs=0;
	    XtSetArg(args[nargs],XmNtitle,ptr); nargs++;
	    XtSetValues(appshell,args,nargs);
	}
	break;
    }
}
/**************************** getuname ************************************/
int getuname(void)
{
#ifdef WIN32
  /* Get system information */
    BOOL status1,status2;
    TCHAR buffer[MAX_COMPUTERNAME_LENGTH+1];
    DWORD nSize=MAX_COMPUTERNAME_LENGTH+1;
    OSVERSIONINFO versioninfo;
    LPSTR platform;

    versioninfo.dwOSVersionInfoSize=sizeof(OSVERSIONINFO);
    status1=GetVersionEx(&versioninfo);
    if(status1) {
	switch(versioninfo.dwPlatformId) {
	case VER_PLATFORM_WIN32s:
	    platform="Windows 3.1";
	    break;
	case VER_PLATFORM_WIN32_WINDOWS:
	    if(versioninfo.dwMajorVersion > 4 ||
	      (versioninfo.dwMajorVersion == 4 && versioninfo.dwMinorVersion)) {
		platform="Windows 98";
	    } else {
		platform="Windows 95";
	    }
	    break;
	case VER_PLATFORM_WIN32_NT:
	    platform="Windows NT";
	    break;
	}
    } else {
	platform="Unknown OS";
    }
    status2=GetComputerName(buffer,&nSize);

    unamestring=calloc(1024,sizeof(char));     /* Danger, fixed size */
    if(!unamestring) {
	xerrmsg("Could not allocate space for system information");
	unamestring=(char *)0;
	return 0;
    }

    sprintf(unamestring,"Running %s %d.%02d on %s",
      platform,
      status1?versioninfo.dwMajorVersion:0,
      status1?versioninfo.dwMinorVersion:0,
      status2?buffer:"Unknown");

    return 1;    
#else
  /* Get system info from uname */
    struct utsname name;
    int len=0;
    
    if(uname(&name) == -1) {
	xerrmsg("Could not find system information");
	unamestring=(char *)0;
	return 0;
    }
    len+=strlen(name.sysname);
    len+=strlen(name.release);
    len+=strlen(name.nodename);
/*     len+=strlen(name.version); */
/*     len+=strlen(name.machine); */
    len+=14;
    unamestring=calloc(len,sizeof(char));
    if(!unamestring) {
	xerrmsg("Could not allocate space for system information");
	unamestring=(char *)0;
	return 0;
    }
    sprintf(unamestring,"Running %s %s on %s",
      name.sysname,name.release,name.nodename);

    return 1;
#endif    
}
/**************************** textcb **************************************/
void textcb(Widget w, XtPointer clientdata, XtPointer calldata)
{
#ifdef TEST
    int n;
    n=(int)clientdata;

    switch(n) {
    case 0:
	printf("textcb\n");
	break;
    }
#endif
}
/**************************** helpcb **************************************/
void helpcb(Widget w, XtPointer clientdata, XtPointer calldata)
{
    Widget w1,wparent;
    int n;
    n=(int)clientdata;

    switch(n) {
    case 0:
	nargs=0;
	XtSetArg(args[nargs],XmNdialogTitle,
	  XmStringCreate("Help",XmSTRING_DEFAULT_CHARSET)); nargs++;
	helppopup=XmCreateFormDialog(mainwin,"optPopup",args,nargs);
	
	nargs=0;
	XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_FORM); nargs++;
	XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
	w1=XmCreatePushButton(helppopup,"Remove",args,nargs);
	XtManageChild(w1);
	XtAddCallback(w1,XmNactivateCallback,helpcb,(XtPointer)2);

	wparent=w1;
	nargs=0;
	XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_WIDGET); nargs++;
	XtSetArg(args[nargs],XmNtopWidget,wparent); nargs++;
	XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
	XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
	XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_FORM); nargs++;
	XtSetArg(args[nargs],XmNeditMode,XmMULTI_LINE_EDIT); nargs++;
	XtSetArg(args[nargs],XmNeditable,False); nargs++;
	XtSetArg(args[nargs],XmNcursorPositionVisible,False); nargs++;
	XtSetArg(args[nargs],XmNrows,25); nargs++;
	XtSetArg(args[nargs],XmNcolumns,80); nargs++;
	w1=XmCreateScrolledText(helppopup,"scrolledHelp",args,nargs);
        XmTextSetString(w1,helptext);
	XtManageChild(w1);

	XtManageChild(helppopup);
	break;
    case 1:
	xinfomsg("%s\n%s",NAMECAPTURE_VERSION_STRING,unamestring);
	break;
    case 2:
	XtDestroyWidget(helppopup);
	break;
    }
}
/**************************** killwidget **********************************/
void killwidget(Widget w, XtPointer clientdata, XtPointer calldata)
{
    XtDestroyWidget(w);
}
/**************************** quit ****************************************/
void quit(Widget w, XtPointer clientdata, XtPointer calldata)
{
    XtDestroyWidget(appshell);
    exit(0);
}
/**************************** timerproc ***********************************/
void timerproc(XtPointer clientdata, XtIntervalId *id)
{
    XmTextPosition len;
    static int limit=0;

  /* Select the new text */
    len=XmTextGetLastPosition(text);
  /*    printf("Timerproc Last Position: |%d|\n",len); */
    if(!len) {
	if(limit++ < 20)
	  XtAppAddTimeOut(appcontext,200,timerproc,NULL);
	else limit=0;     /* For next time */
	return;
    }
    limit=0;
    XmTextSetSelection(text,0,len,CurrentTime);
    if(file && autofile) {
	filecb((Widget)0,(XtPointer)3,(XtPointer)0);
    }
}
/**************************** workproc ************************************/
#ifdef TEST
Boolean workproc(XtPointer clientdata)
{
    XmTextPosition len;

  /* Select the new text */
    len=XmTextGetLastPosition(text);
  /*    printf("Workproc Last Position: |%d|\n",len); */
    XmTextSetSelection(text,0,len,CurrentTime);
    return(True);          /* Don't call work proc again */
}
#endif
/**************************** writeburtheader *****************************/
void writeburtheader(FILE *file)
{
    fprintf(file,"%s\n",SDDSID);
    fprintf(file,
      "&description text=\"Namecapture BURT Request File\""
      " contents=\"BURT Request\" &end\n");
    fprintf(file,
      "&column name=ControlName type=string &end\n"
      "&column name=ControlType type=string &end\n"
      "&column name=ControlMode type=string &end\n"
      "&data mode=\"ascii\" no_row_counts=1 additional_header_lines=1 &end\n"
      "\n");
}
/**************************** xerrmsg *************************************/
void xerrmsg(char *fmt, ...)
{
    Widget child;
    va_list vargs;
    static char lstring[BUFSIZ];
    
    va_start(vargs,fmt);
    (void)vsprintf(lstring,fmt,vargs);
    va_end(vargs);
    
    if(lstring[0] != '\0') {
	if(windowmessage) {
	    XBell(display,50); XBell(display,50); XBell(display,50); 
	    cstring=XmStringCreateLtoR(lstring,XmSTRING_DEFAULT_CHARSET);
	    nargs=0;
	    XtSetArg(args[nargs],XmNtitle,"Warning"); nargs++;
	    XtSetArg(args[nargs],XmNmessageString,cstring); nargs++;
	    warningbox=XmCreateWarningDialog(mainwin,"warningMessage",
	      args,nargs);
	    XmStringFree(cstring);
	    child=XmMessageBoxGetChild(warningbox,XmDIALOG_CANCEL_BUTTON);
	    XtDestroyWidget(child);
	    child=XmMessageBoxGetChild(warningbox,XmDIALOG_HELP_BUTTON);
	    XtDestroyWidget(child);
	    XtManageChild(warningbox);
	    XtAddCallback(warningbox,XmNokCallback,killwidget,NULL);
	}
#ifdef WIN32
	lprintf("%s\n",lstring);
#else
	fprintf(stderr,"%s\n",lstring);
#endif
    }
}
/**************************** xinfomsg ************************************/
void xinfomsg(char *fmt, ...)
{
    Widget child;
    va_list vargs;
    static char lstring[BUFSIZ];
    
    va_start(vargs,fmt);
    (void)vsprintf(lstring,fmt,vargs);
    va_end(vargs);
    
    if(lstring[0] != '\0') {
	cstring=XmStringCreateLtoR(lstring,XmSTRING_DEFAULT_CHARSET);
	nargs=0;
	XtSetArg(args[nargs],XmNtitle,"Information"); nargs++;
	XtSetArg(args[nargs],XmNmessageString,cstring); nargs++;
	infobox=XmCreateInformationDialog(mainwin,"informationMessage",
	  args,nargs);
	XmStringFree(cstring);
	child=XmMessageBoxGetChild(infobox,XmDIALOG_CANCEL_BUTTON);
	XtDestroyWidget(child);
	child=XmMessageBoxGetChild(infobox,XmDIALOG_HELP_BUTTON);
	XtDestroyWidget(child);
	XtManageChild(infobox);
	XtAddCallback(infobox,XmNokCallback,killwidget,NULL);
#ifdef WIN32
	lprintf("%s\n",lstring);
#else
	fprintf(stderr,"%s\n",lstring);
#endif
    }
}
/**************************** yesno ***************************************/
int yesno(Widget wparent, char * question)
  /* Based on O'Reilly */
  /* Returns 1 for yes, 0 for no, -1 for cancel */
{
    static Widget wdialog=(Widget)0;
    XmString message,yes,no,cancel;
    static int answer;
    
    if (!wdialog) {
        yes=XmStringCreateLocalized("Yes");
        no=XmStringCreateLocalized("No");
        cancel=XmStringCreateLocalized("Cancel");
	nargs=0;
	XtSetArg(args[nargs],XmNokLabelString,yes); nargs++;
	XtSetArg(args[nargs],XmNcancelLabelString,no); nargs++;
	XtSetArg(args[nargs],XmNhelpLabelString,cancel); nargs++;
	XtSetArg(args[nargs],XmNdialogStyle,XmDIALOG_FULL_APPLICATION_MODAL); nargs++;
	wdialog=XmCreateQuestionDialog(wparent,"yesNo",args,nargs);
        XtAddCallback(wdialog,XmNokCallback,yesnoresponse,&answer);
        XtAddCallback(wdialog,XmNcancelCallback,yesnoresponse,&answer);
        XtAddCallback(wdialog,XmNhelpCallback,yesnoresponse,&answer);
	XmStringFree(yes);
	XmStringFree(no);
	XmStringFree(cancel);

      /* Set the title of the dialog shell. (This is what the user sees.)
	 Otherwise it is named yesNo_popup.  */
	nargs=0;
	XtSetArg(args[nargs],XmNtitle,"Question"); nargs++;
	XtSetValues(XtParent(wdialog),args,nargs);
    }
    message=XmStringCreateLocalized(question);
    nargs=0;
    XtSetArg(args[nargs],XmNmessageString,message); nargs++;
    XtSetValues(wdialog,args,nargs);
    XmStringFree(message);
    XtManageChild(wdialog);
    XtPopup(XtParent(wdialog),XtGrabNone);
  /* Loop until user decides what to do */
    answer=-2;
    while (answer == -2 || XtAppPending(appcontext))
      XtAppProcessEvent(appcontext,XtIMAll);
    XtPopdown(XtParent(wdialog));
    XSync(display,0);
    XmUpdateDisplay(wparent);
    return answer;
}
/**************************** yesnoresponse *******************************/
void yesnoresponse(Widget w, XtPointer clientdata, XtPointer calldata)
{
    int *answer=(int *)clientdata;
    XmAnyCallbackStruct *cbs=(XmAnyCallbackStruct *)calldata;
    
    switch (cbs->reason) {
    case XmCR_OK:
	*answer=1;
	break;
    case XmCR_CANCEL:
	*answer=0;
	break;
    case XmCR_HELP:
	*answer=-1;
	break;
    default:
	break;
    }
}
